#include <GL/glut.h>
#include <pthread.h>

#include "../graphics/graphics.h"
#include "gl2ps.h"

extern int winw, winh;
extern const double vmarg;

void setProjectionMatrix ();

typedef enum _my_fonts {myfont_BITMAP_8_BY_13 = 0, 
			myfont_BITMAP_9_BY_15 = 1, 
			myfont_BITMAP_TIMES_ROMAN_10 = 2, 
			myfont_BITMAP_TIMES_ROMAN_24 = 3, 
			myfont_BITMAP_HELVETICA_10 = 4, 
			myfont_BITMAP_HELVETICA_12 = 5, 
			myfont_BITMAP_HELVETICA_18 = 6} _my_fonts;

void *gl_font[] = {GLUT_BITMAP_8_BY_13, GLUT_BITMAP_9_BY_15, GLUT_BITMAP_TIMES_ROMAN_10, GLUT_BITMAP_TIMES_ROMAN_24, GLUT_BITMAP_HELVETICA_10, GLUT_BITMAP_HELVETICA_12, GLUT_BITMAP_HELVETICA_18};

char *gl2ps_font[] = {"Fixed","Fixed","TimesRoman","TimesRoman","Helvetica","Helvetica","Helvetica"};

int fontsize[] = {13, 15, 10, 24, 10, 12, 18};


void drawstring(const char str[], _my_fonts f) {
  const char *c = str;

  gl2psText(str, gl2ps_font[f], fontsize[f]);

  while (*c!='\0') {
    glutBitmapCharacter(gl_font[f],*c++);
  }
}

extern GLfloat realzoom;

int drawcall(const _grcall **tmp_grcall, const _grcall * const last_of_stack) {
  /* _line_elem *line; */
  static double cpos[3] = {0.};
  _my_fonts f = myfont_BITMAP_HELVETICA_12;
  double strln;
  double strht;
  double locsc = realzoom*1.6/(double)winw;
  char *str;
  double *coord, *val;
  int t, col;
  const double rectsize = .015;

  t = (*tmp_grcall)->type;
  val = (*tmp_grcall)->val;
  switch (t) {
  case CLEAR:
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    break;
  case EXIT:
    pthread_exit(NULL);
    break;
  case TWOD:
    setProjectionMatrix();
    break;
  case THREED:
    setProjectionMatrix();
    break;
  case MOVE:
    coord = (double*)val;
    cpos[0] = coord[0];
    cpos[1] = coord[1];
    cpos[2] = coord[2];
    break;
  case GR_POINT:
    coord = (double*)val;
    glBegin(GL_POINTS);
    {
      glVertex3dv((double*)val);
    }
    glEnd();
    /*  remember last position of the line */
    cpos[0] = coord[0];
    cpos[1] = coord[1];
    cpos[2] = coord[2];
    break;
  case FRECT:
  case ORECT:
    coord = (double*)val;
    if (t == FRECT)
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    else
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glBegin(GL_POLYGON);
    {
      glVertex3d(coord[0]-rectsize, coord[1]-rectsize, coord[2]-rectsize);
      glVertex3d(coord[0]-rectsize, coord[1]+rectsize, coord[2]-rectsize);
      glVertex3d(coord[0]+rectsize, coord[1]+rectsize, coord[2]-rectsize);
      glVertex3d(coord[0]+rectsize, coord[1]-rectsize, coord[2]-rectsize);
    }
    glEnd();
    /*  remember last position */
    cpos[0] = coord[0];
    cpos[1] = coord[1];
    cpos[2] = coord[2];
    break;
  case DRAW:
    glBegin(GL_LINE_STRIP);
    {
      glVertex3dv(cpos);
      coord = (double*)val;
      glVertex3dv(coord);
      /*  optimizing drawing of lines: */
      while (*tmp_grcall != last_of_stack && 
	     (*tmp_grcall)->next->type == DRAW) {
	*tmp_grcall = (*tmp_grcall)->next;
	coord = (double*)((*tmp_grcall)->val);
	glVertex3dv(coord);
      }
      /*  remember last position of the line */
      cpos[0] = coord[0];
      cpos[1] = coord[1];
      cpos[2] = coord[2];
    }
    glEnd();
    break;
  case TEXT:
    str = (char*)val;
    strln = locsc * glutBitmapLength(gl_font[f], (const unsigned char*)str);
    strht = locsc * fontsize[f];

    glRasterPos3d(cpos[0] - strln/2.,cpos[1] - strht/2., cpos[2]);
    drawstring(str, f);
    break;
  case LTEXT:
    str = (char*)val;
    strht = locsc * fontsize[f];

    glRasterPos3d(cpos[0],cpos[1], cpos[2]);
    drawstring(str, f);
    break;
  case RTEXT:
    str = (char*)val;
    strln = locsc * glutBitmapLength(gl_font[f], (const unsigned char*)str);
    strht = locsc * fontsize[f];

    glRasterPos3d(cpos[0] - strln,cpos[1] - strht/2., cpos[2]);
    drawstring(str, f);
    break;
  case COLOR:
    col = colorWrap(*((int*)val));
    glColor3dv(colormap[col]);
    break;
  default:
    fprintf(stderr,"error: trying to draw unknown primitive: %d\n", t);
    break;
  }
  return 1;
}

int ExponentText(char *base, char *pow, double x, double y, double z) {
  addGrcall(MOVE, NULL, x, y, z);
  addGrcall(RTEXT, base, 0., 0., 0.);
  addGrcall(MOVE, NULL, x, y, z);
  addGrcall(LTEXT, pow, 0., 0., 0.);
  return 1;
}
