#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "pp.h"

#define TINY 1e-12
#define CLEAN 1
#define DEBUG 0

extern NODEP *dffeqns, *asignms;
extern int jacob;
extern NAMEP linlist;
extern char **varnms, **ivrnms;
extern int nvar, nimvar, nnames;
extern char warnings[21][80];
extern int nwarning, nlines;

NODEP deriv(),clean();

int isint(s)
char s[];
{
  int i,n;
  for (n=1,i=0; s[i] != '\0' && n; i++)
    if (!(isdigit((int)(s[i])) || s[i] == '+' || s[i] == '-')) n = 0;
  return (n);
}

int mkfloat(s)
char s[];
{
  int i;
  for (i=0; s[i] != '\0'; i++)
    if (s[i] == '+' || s[i] == '-' || 
	s[i] == '.' || s[i] == 'd' || s[i] == 'e') return(0);
  s[i] = '.';
  s[i+1] = '\0';
  return(i+1);
}

int equal_floats(a,b)
float a,b;
{
   return (a-b < TINY && b-a < TINY);
}

NODEP mknode(opr,text,left,right)
char opr,text[];
NODEP left, right;
{
   NODEP node;
   node = (NODEP)malloc((unsigned)sizeof(NODE));
   if (node == NULL) {printf("PP GRIND: out of memory\n"); exit(0);}
   node->opr = opr;
   node->left = left;
   node->right = right;
   strncpy(node->text, text, 11);  
   if (opr == '1' && isint(text)) mkfloat(node->text);
   return(node);
}

NODEP cpnode(node,opr,text,left,right)
char opr,text[];
NODEP node,left, right;
{
   node->opr = opr;
   strncpy(node->text, text,11);
   node->left = left;
   node->right = right;
   return(node);
}

NODEP cptree(node)
NODEP node;
{
  if (node != NULL)
    return (mknode(node->opr, node->text,
            cptree(node->left),cptree(node->right)));
  else return(NULL);
}

NODEP rmtree(node)
NODEP node;
{
  if (node != NULL) {
    rmtree(node->left);
    rmtree(node->right);
    free(node);
  }
  return(NULL);
}

int prtree(node,f)
NODEP node;
FILE *f;
{
  int narg = 0, func = 0;
  if (node != NULL) {
    if (node->right != NULL) narg = 1;
    if (node->left != NULL) narg = 2;
    if (node->opr == 'f') func = 1;
    if (node->opr == 'g') func = 2;
    if (!func) {
      if (narg) putc('(',f);
      if (narg == 2) {prtree(node->left,f);}
      if (node->opr == '1' && node->text[0] == '-') fprintf(f,"(%s)",node->text);
      else if ('0' == node->opr)  fprintf(f,"%s_",node->text);
      else fprintf(f,"%s",node->text);
      if (narg) prtree(node->right,f);
      if (narg) putc(')',f);
    } else {
      fprintf(f,"%s",node->text);
      putc('(',f);
      if (func == 2) {
        prtree(node->left,f);
        putc(',',f);
        prtree(node->right,f);
      } else {
        prtree(node->right,f);
      }
      putc(')',f);
    }
  }
  return 1;
}

int eqtree(node1,node2)
NODEP node1, node2;
{
  if (node1 == NULL && node2 == NULL)
    return(1);
  else {
    if (node1 == NULL || node2 == NULL)
      return(0);
    else {
      if ((strcmp(node1->text,node2->text) == 0) &&
          eqtree(node1->left,node2->left) &&
          eqtree(node1->right,node2->right) )
        return(1);
    }
  }
  return(0);
}

NODEP pasteim(node, i)
NODEP node; int i;
{
  if (node != NULL) {
    if ((strcmp(node->text,ivrnms[i]) == 0))
      node = cptree(asignms[i]);
    else {
      node->left  = pasteim(node->left, i);
      node->right = pasteim(node->right,i);
    }
  }
  return(node);
}

NODEP cutim(node,i)
NODEP node; int i;
{
  if (node != NULL) {
    if (eqtree(node,asignms[i]))
      node = mknode('0',ivrnms[i],NULL,NULL);
    else {
      node->left  = cutim(node->left, i);
      node->right = cutim(node->right,i);
    }
  }
  return(node);
}

NODEP clean(node)
NODEP node;
{
  int left_num, right_num;
  double left_value, right_value, value, value2;
  NODEP left,right;
  char text[11];
  strncpy(text,"none",11);
  if (node != NULL && CLEAN) {
    left = node->left;
    right = node->right;
    if ( (left_num = (left != NULL && left->opr == '1')) )
      sscanf(left->text,"%lf",&left_value);
    if ( (right_num = (right != NULL && right->opr == '1')) )
      sscanf(right->text,"%lf",&right_value);

    switch (node->opr) {
    case '+' : 
      if (left_num && right_num) {
	sprintf(text,"%-g", value = left_value+right_value);
	if (isint(text)) mkfloat(text);
        sscanf(text,"%lf",&value2);
	if (equal_floats(value,value2)) {
	  if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,text);
	  rmtree(left); rmtree(right);
	  return(cpnode(node,'1',text,NULL,NULL));
	} else if (DEBUG) printf("FORMAT: %s %f %f\n",text,value,value2);
      }
      if (left_num && equal_floats(left_value,0.0)) {
        if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,right->text);
	rmtree(left); 
	return(right);
      }
      if (right_num && equal_floats(right_value,0.0) && left != NULL) {
        if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,left->text);
	rmtree(right);
	return(left);
      }
      break;
    case '-' :
      if (left_num && right_num && left_value >= right_value) {
	sprintf(text,"%-g", value = left_value-right_value);
	if (isint(text)) mkfloat(text);
        sscanf(text,"%lf",&value2); 
	if (equal_floats(value,value2)) {
	  if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,text);
	  rmtree(left); rmtree(right);
	  return(cpnode(node,'1',text,NULL,NULL));
	} else if (DEBUG) printf("FORMAT: %s %f %f\n",text,value,value2);
      }
      if (right_num && equal_floats(right_value,0.0) && left != NULL) {
	if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,left->text);
	rmtree(right); 
	return(left);
      }
      if (right_num && equal_floats(right_value,0.0) && left == NULL) {
	if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,"0");
        rmtree(right);
        return(cpnode(node,'1',"0.",NULL,NULL));
      }
      break;
    case '*' : 
      if ((left_num && equal_floats(left_value,0.0)) || 
	  (right_num && equal_floats(right_value,0.0))) {
	if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,"0");
        rmtree(left); rmtree(right);
        return(cpnode(node,'1',"0.",NULL,NULL));
      }
      if (left_num && equal_floats(left_value,1.0)) {
	if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,right->text);
	rmtree(left); 
	return(right);
      }
      if (right_num && equal_floats(right_value,1.0)) {
	if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,left->text);
	rmtree(right); 
	return(left);
      }
      break;
    case '/' : 
      if (left_num && equal_floats(left_value,0.0)) {
	if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,"0");
        rmtree(left); rmtree(right);
        return(cpnode(node,'1',"0.",NULL,NULL));
      }
      if (right_num && equal_floats(right_value,1.0)) {
	if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,left->text);
	rmtree(right); 
	return(left);
      }
      if (eqtree(left,right)) {
	if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,"1");
        rmtree(left); rmtree(right);
        return(cpnode(node,'1',"1.",NULL,NULL));
      }
      break;
    case 'g' : 
      if (0 == strcmp(node->text,"pow")){
	if (right_num && equal_floats(right_value,0.0)) {
	  if (DEBUG) printf("Hi %s %s %s => %s\n",left->text,node->text,right->text,"1");
	  rmtree(left); rmtree(right);
	  return(cpnode(node,'1',"1.",NULL,NULL));
	}
	if (right_num && equal_floats(right_value,1.0)) {
	  if (DEBUG) printf("%s %s %s => %s\n",left->text,node->text,right->text,left->text);
	  rmtree(right); 
	  return(left);
	}
      }
      break;
    case 'f':
      if (0 == strcmp(node->text,"sqrt") && 0 == strcmp(right->text,"pow") 
	  && 0 == strcmp(right->right->text,"2.")) {
	if (DEBUG) {prtree(right->left,stderr);fprintf(stderr,"\n");}
	return cpnode(node,right->left->opr, right->left->text, 
		      right->left->left, right->left->right);
      }
      break;
    default : ;
    }
  }
  return(node);
}

NODEP cltree(node)
NODEP node;
{
  if (node != NULL) {
    node->left = cltree(node->left);
    node->right = cltree(node->right);
    return (clean(node));
  }
  else return(node);
}

NODEP deriv(node,text)
NODEP node; char text[];
{
  NODEP left,right,dleft,dright;
  NODEP l1,l2,r1,t1,t2;
  if (node != NULL && jacob) {
    /* node = cltree(node); */
    left = node->left;
    right = node->right;
    switch (node->opr) {
    case '0' : 
      if ((strcmp(node->text,text)==0))
           return(cpnode(node,'1',"1.",NULL,NULL));
      else return(cpnode(node,'1',"0.",NULL,NULL));
    case '1' : 
      return(cpnode(node,'1',"0.",NULL,NULL));
    case '+' : 
      return(clean(cpnode(node,'+',"+",deriv(left,text),deriv(right,text))));
    case '-' : 
      return(clean(cpnode(node,'-',"-",deriv(left,text),deriv(right,text))));
    case '*' : 
      dleft = deriv(cptree(left),text);
      dright = deriv(cptree(right),text);
      return(clean(cpnode(node,'+',"+",
           clean(mknode('*',"*",dleft,right)),
           clean(mknode('*',"*",left,dright)))));
    case '/' : 
      dleft = deriv(cptree(left),text);
      dright = deriv(cptree(right),text);
      t1 = mknode('*',"*",cptree(right),cptree(right));
      t2 = mknode('-',"-",clean(mknode('*',"*",dleft,right)),
                          clean(mknode('*',"*",left,dright)));
      return(clean(cpnode(node,'/',"/",t2,t1)));
    case 'f' : 
      dright = deriv(cptree(right),text);
      if (strcmp(node->text,"exp") == 0)
        return(clean(cpnode(node,'*',"*",dright,
          mknode('f',"exp",left,right))));
      if (strcmp(node->text,"sin") == 0)
        return(clean(cpnode(node,'*',"*",dright,
          mknode('f',"cos",left,right))));
      if (strcmp(node->text,"cos") == 0)
        return(clean(cpnode(node,'*',"*",dright,
          mknode('-',"-",mknode('1',"0.",NULL,NULL),mknode('f',"sin",left,right)))));
      if (strcmp(node->text,"sqrt") == 0)
        return(clean(mknode('/',"/",dright,
          mknode('*',"*",mknode('1',"2.",NULL,NULL),node))));
      if (strcmp(node->text,"log") == 0)
        return(clean(cpnode(node,'*',"*",dright,
          mknode('/',"/",mknode('1',"1.",NULL,NULL),right))));
      if (strcmp(node->text,"tan") == 0)
        return(clean(mknode('*',"*",dright,
          mknode('+',"+",mknode('1',"1.",NULL,NULL),
            mknode('g',"pow",node,mknode('1',"2.",NULL,NULL))))));
      if (strcmp(node->text,"tanh") == 0)
        return(clean(mknode('*',"*",dright,
          mknode('-',"-",mknode('1',"1.",NULL,NULL),
            mknode('g',"pow",node,mknode('1',"2.",NULL,NULL))))));
      printf("\nDon't know how to derive: %s\n",node->text);
      jacob = 0;
      return(node);
    case 'g' : 
      if (strcmp(node->text,"pow") == 0) {
	dleft = deriv(cptree(left),text);
	dright = deriv(cptree(right),text);
	l1 = cptree(left);
	r1 = cptree(right);
	if (dright->text[0] == '0') {
	  t1 = clean(mknode('-',"-",r1,mknode('1',"1.",NULL,NULL)));
	  t2 = clean(mknode('*',"*",cptree(right),clean(cpnode(node,'g',"pow",l1,t1))));
	  return(clean(mknode('*',"*",t2,dleft)));
	}else{
	  l2 = cptree(left);
	  t1 = clean(mknode('*',"*",mknode('f',"log",NULL,l1),dright));
	  t2 = clean(mknode('/',"/",mknode('*',"*",r1,dleft),l2));
	  return(clean(mknode('*',"*",node,mknode('+',"+",t1,t2))));
	}
      }
      printf("\nDon't know how to derive: %s\n",node->text);
      jacob = 0;
      return(node);
    default  :
      printf("\nUnknown operator: %s\n",node->text);
      return(node);
    
    }
  }
  return(node);
}

