#include <stdlib.h>
#include <string.h>

#include "grind.h"

double finish=100.;
int nintervals = 10;
int itor = ROW4A;
double ztol = 1.e-6;
int ngrun = WHITE;

extern boolean trunc_opt;
extern boolean noiyes_opt;
extern double delta;
extern int rundatalength;
extern double **rundata;

int checkRunBuffer(int);

/* int dnorm(int a,double b[],double c[]) {return 1;} */

int onestep(double t, double tend, double pop[]) {
  /*  perform one integration step, from t to tend */
  /*  variable values are stored in pop[], and the results will be stored */
  /*  in pop[] as well */

  /*  'tend' was a regular double */
  double hmin = 0.0, hi = 1.e-3;
  double hmax = (tend-t)/60.0;
  static double *dy = NULL; /*  a dummy, only used in call to fcn() */
  int nok,nbad; /*  dummy var for odeint; stores number of ok and bad trials. */

  int ipas = 100000, irep;

  if (Interrupt()) return 0;
  if (dy == NULL) 
    dy = dcalloc(nvar, "'dy' in onestep()");
  
  if (map_opt) {
    itor = EULER;
    delta = 1.0;
  }
  if (tend < TINY)
    fcn(pop, dy);
  else {
    switch (itor) {
    case ROW4A:
      row4a(nvar, &t, pop, tend, ztol, hmax, hi, &ipas, &irep);
      break;
    case RK:
      odeint((pop-1),nvar,t,tend,ztol,hi,hmin,&nok,&nbad,nfcn,rkqs);
      break;
    case EULER:
      euler(pop, nvar, t, tend);
      break;
    default:
      gr_warning("unknown integrator in onestep(); skipping integration\n");
      return 0;
      break;
    }
  }
  return !Interrupt();
}

int report(FILE *datafile, double t, int istep) {
  int v;
  static double *dy = NULL;

  if (Interrupt()) return 0;

  if (dy == NULL) 
    dy = dcalloc(nvar, "'dy' in integr()");

  /*  check whether we have a buffer to write to */
  if (rundatalength) {
    /*  update rundata buffer */
    rundata[0][istep]=t;
    for (v=0; v<nnames; v++)
      if (outputbuf[v])
	rundata[v+1][istep] = all_values[v];
  }
  /*  if file open, write to it */
  if (datafile) {
    fprintf(datafile, dformat, t); 
    for (v=0; v<nnames; v++)
      if (outputbuf[v])
	fprintf(datafile, dformat, all_values[v]);
    fprintf(datafile,"\n");
    fflush(datafile);
  }
  /*  if we have 2d/3d graphical mode */
  if (mode_opt >=2) {
    /*  update graphical data */
    if (istep == 0) {
      if (initial_opt) {
	fcn(fstate,dy);
	triangle(fstate, dy);
      }
      start_trajectory(ngrun);
    }
    else
      add2trajectory();
  }
  /*  if alpha-num mode */
  else if (mode_opt == 1) {
    printf(dformat, t);
    for (v=0; v<nnames; v++)
      if (outputbuf[v])
	printf(dformat, all_values[v]);
    printf("\n");
  }
  return 1;
}

/*  ibuf has length nvar+1 but will be a dummy char */
int integr(FILE *datafile) {
  int p;
  double t=0.0,tend;
  int nstep, istep;

  if (map_opt) { 
    itor = EULER; 
    delta = 1.0; 
  }
  if (itor == EULER && finish/nintervals < delta) 
    nintervals = finish/delta + 0.5;
  nstep = nintervals;
  checkRunBuffer(nstep);
  for (istep = 0; istep<=nstep; istep++) {
    tend = finish*istep/nstep;
    /*  set noise */
    for (p=0; p<npar; p++) {
      if (gaussian == pars[p].noise)
        par_values[nexpr + p] = znormal(pars[p].value, pars[p].sd, pars[p].pos);
    if (uniform == pars[p].noise)
        par_values[nexpr + p] = zuniform();
    }
    if (onestep(t,tend,fstate))
      report(datafile, tend, istep);
    else
      break;
    if (trunc_opt && outside(all_values[axis[0].var],
			     all_values[axis[1].var],
			     all_values[axis[2].var]))
      break;

    t = tend;
  }
  /*  unset noise */
  for (p=0; p<npar; p++)
    if (pars[p].noise)
      par_values[nexpr + p] = pars[p].value;
  return 1;
}

int checkRunBuffer(int nstep) {
  boolean memerror = false;
  int i;

  if (Interrupt()) return 0;
  /*  check whether time-column of the rundata buffer is appropriate */
  if (rundatalength <= nstep) {
    if (!(rundata[0]= (double*)realloc(rundata[0],(nstep+1)*sizeof(double))))
      memerror = true;
    memset(rundata[0], 0, (nstep+1)*sizeof(double));
  }
  /*  check for all var/expr/par whether output should be generated */
  /*  if so, check and adjust the corresponding column of the rundata buffer */
  /*  if not, free the corresponding column, if it exists */
  for (i=0; i<nnames; i++) {
    if (outputbuf[i]) {
      /*  if buffer doesn't exist or is too small */
      if (!rundata[i+1] || rundatalength <= nstep) {
	if (!(rundata[i+1]=
	      (double*)realloc(rundata[i+1],(nstep+1)*sizeof(double)))) {
	  memerror = true;
	  break;
	}
	/* */
	memset(rundata[0], 0, (nstep+1)*sizeof(double));
	/* */
      }
    }
    /*  if output is not wanted but buffer exists; free it */
    else if (rundata[i+1]) {
      free(rundata[i+1]);
      rundata[i+1] = NULL;
    }
  }
  /*  if an error occurred during memory allocation free everything */
  /*  and set the length of the buffer to zero */
  if (memerror) {
    printf("too much data for saving timeplot buffer\n");
    rundatalength = 0;
    for (i=0; i<1+nnames; i++) {
      if(rundata[i]) 
	free (rundata[i]);
      rundata[i] = NULL;
    }
  }
  else
    /*  set the new length of the rundata buffer */
    rundatalength = nstep + 1;
  return 1;
}

int plot() {
  int nstep, i, n;
  boolean ok2plot = true;
  nstep = nintervals; 

  if (mode_opt <DIM2) {
    gr_warning("not in graphical mode (2d or 3d)\n");
    return 0;
  }

  for (n=0; n<nvar; n++)
    if (NULL == rundata[n+1])
      ok2plot = false;
  if (rundatalength <= nstep)
    ok2plot = false;

  if (!ok2plot) {
    gr_warning("rundata buffer not appropriate for plotting\n");
    return 0;
  }
  
  for (n=0; n<nvar; n++)
    all_values[n] = rundata[n+1][0];
  start_trajectory(++ngrun);
  
  for (i=1; i<nstep; i++) {
    for (n=0; n<nvar; n++)
      all_values[n] = rundata[n+1][i];
    if (Interrupt()) return 0;
    add2trajectory();
    if (trunc_opt && outside(all_values[axis[0].var],
			     all_values[axis[1].var],
			     all_values[axis[2].var]))
      break;

  }
  return 1;
}
